import pytest

from popidd_io import get_anno_reader, get_image_reader

# make_napari_viewer is a pytest fixture that returns a napari viewer object
# you don't need to import it, as long as napari is installed
# in your testing environment

# tmp_path is a pytest fixture

# def test_image_reader(tmp_path):
#     """An example of how you might test your plugin."""

#     # write some fake data using your supported file format
#     my_test_file = str(tmp_path / "myfile.npy")
#     original_data = np.random.rand(20, 20)
#     np.save(my_test_file, original_data)

#     # try to read it back in
#     reader = napari_get_reader(my_test_file)
#     assert callable(reader)

#     # make sure we're delivering the right format
#     layer_data_list = reader(my_test_file)
#     assert isinstance(layer_data_list, list) and len(layer_data_list) > 0
#     layer_data_tuple = layer_data_list[0]
#     assert isinstance(layer_data_tuple, tuple) and len(layer_data_tuple) > 0

#     # make sure it's the same as it started
#     np.testing.assert_allclose(original_data, layer_data_tuple[0])


# def test_get_reader_pass():
#     reader = napari_get_reader("fake.file")
#     assert reader is None


# Add here pass and fail tests for the readers
@pytest.fixture
def write_im2file(tmp_path):
    def write_tiff(file):
        import numpy
        from tifffile import TiffWriter

        # TIff generation snippet from tiffile
        data = numpy.random.randint(0, 255, (8, 2, 512, 512, 3), "uint8")
        subresolutions = 2
        pixelsize = 0.29  # micrometer
        tiff_file = f"{tmp_path}/{file}"
        with TiffWriter(tiff_file, bigtiff=True) as tif:
            metadata = {
                "axes": "TCYXS",
                "SignificantBits": 8,
                "TimeIncrement": 0.1,
                "TimeIncrementUnit": "s",
                "PhysicalSizeX": pixelsize,
                "PhysicalSizeXUnit": "µm",
                "PhysicalSizeY": pixelsize,
                "PhysicalSizeYUnit": "µm",
                "Channel": {"Name": ["Channel 1", "Channel 2"]},
                "Plane": {
                    "PositionX": [0.0] * 16,
                    "PositionXUnit": ["µm"] * 16,
                },
                "Description": "A multi-dimensional, multi-resolution image",
                "MapAnnotation": {  # for OMERO
                    "Namespace": "openmicroscopy.org/PyramidResolution",
                    "1": "256 256",
                    "2": "128 128",
                },
            }
            options = dict(
                photometric="rgb",
                tile=(128, 128),
                compression="jpeg",
                resolutionunit="CENTIMETER",
                maxworkers=2,
            )
            tif.write(
                data,
                subifds=subresolutions,
                resolution=(1e4 / pixelsize, 1e4 / pixelsize),
                metadata=metadata,
                **options,
            )
            # write pyramid levels to the two subifds
            # in production use resampling to generate sub-resolution images
            for level in range(subresolutions):
                mag = 2 ** (level + 1)
                tif.write(
                    data[..., ::mag, ::mag, :],
                    subfiletype=1,
                    resolution=(1e4 / mag / pixelsize, 1e4 / mag / pixelsize),
                    **options,
                )
            # add a thumbnail image as a separate series
            # it is recognized by QuPath as an associated image
            thumbnail = (data[0, 0, ::8, ::8] >> 2).astype("uint8")
            tif.write(thumbnail, metadata={"Name": "thumbnail"})

        return data, tiff_file

    return write_tiff


def test_img_reader_pass():
    reader = get_image_reader("fake.file")
    assert reader is None


def test_img_reader_list(write_im2file):
    _, path1 = write_im2file("temp1.ome.tif")
    _, path2 = write_im2file("temp2.ome.tif")
    reader = get_image_reader([path1, path2])
    assert callable(reader)


def test_img_reader(write_im2file):
    data, test_file = write_im2file("temp.ome.tif")
    reader = get_image_reader(test_file)
    assert callable(reader)

    layer_data_list = reader(test_file)
    assert isinstance(layer_data_list, list) and len(layer_data_list) > 0
    assert (
        isinstance(layer_data_list[0], tuple) and len(layer_data_list[0]) > 0
    )
    assert data.shape == layer_data_list[0][0][0].shape
    # numpy.testing.assert_allclose(data, layer_data_list[0][0][0].compute()) #Numpy of 1st lvl of dask array list


def test_anno_reader_pass():
    reader = get_anno_reader("fake.file")
    assert reader is None



"""
This module is an example of a barebones numpy reader plugin for napari.

It implements the Reader specification, but your plugin may choose to
implement multiple readers or even other plugin contributions. see:
https://napari.org/stable/plugins/guides.html?#readers
"""

from collections.abc import Callable, Sequence

from ._anno import load_geojson, load_parquet
from ._image import load_img


def get_image_reader(path: str | Sequence[str]) -> Callable | None:
    img_formats = (".tiff", ".tif", ".svs", ".ndpi", ".qptiff")
    if isinstance(path, str):
        if not path.lower().endswith(img_formats):
            return None
    elif isinstance(path, Sequence):
        print(path)
        print(type(path))
        if not all(p.lower().endswith(img_formats) for p in path):
            return None
    else:
        return None
    return load_img


def get_anno_reader(path: str | Sequence[str]) -> Callable | None:
    if path.lower().endswith(".geojson"):
        return load_geojson
    elif path.lower().endswith(".parquet"):
        return load_parquet
    else:
        return None


# def napari_get_reader(path):
#     """A basic implementation of a Reader contribution.

#     Parameters
#     ----------
#     path : str or list of str
#         Path to file, or list of paths.

#     Returns
#     -------
#     function or None
#         If the path is a recognized format, return a function that accepts the
#         same path or list of paths, and returns a list of layer data tuples.
#     """
#     if isinstance(path, list):
#         # reader plugins may be handed single path, or a list of paths.
#         # if it is a list, it is assumed to be an image stack...
#         # so we are only going to look at the first file.
#         path = path[0]

#     # if we know we cannot read the file, we immediately return None.
#     if not path.endswith(".npy"):
#         return None

#     # otherwise we return the *function* that can read ``path``.
#     return reader_function


# def reader_function(path):
#     """Take a path or list of paths and return a list of LayerData tuples.

#     Readers are expected to return data as a list of tuples, where each tuple
#     is (data, [add_kwargs, [layer_type]]), "add_kwargs" and "layer_type" are
#     both optional.

#     Parameters
#     ----------
#     path : str or list of str
#         Path to file, or list of paths.

#     Returns
#     -------
#     layer_data : list of tuples
#         A list of LayerData tuples where each tuple in the list contains
#         (data, metadata, layer_type), where data is a numpy array, metadata is
#         a dict of keyword arguments for the corresponding viewer.add_* method
#         in napari, and layer_type is a lower-case string naming the type of
#         layer. Both "meta", and "layer_type" are optional. napari will
#         default to layer_type=="image" if not provided
#     """
#     # handle both a string and a list of strings
#     paths = [path] if isinstance(path, str) else path
#     # load all files into array
#     arrays = [np.load(_path) for _path in paths]
#     # stack arrays into single array
#     data = np.squeeze(np.stack(arrays))

#     # optional kwargs for the corresponding viewer.add_* method
#     add_kwargs = {}

#     layer_type = "image"  # optional, default is "image"
#     return [(data, add_kwargs, layer_type)]



import pathlib

import geopandas
import pandas
from napari.types import LayerDataTuple

# Add here the geojson  and parquet reading and writing support for qupath compatibility

# geojson to shapes and shapes to geojson

# parquet to shapes for cell annotations


def load_geojson(path: str | pathlib.Path) -> list[LayerDataTuple]:
    """
    Load a GeoJSON file and convert it to a list of shape layers.

    Parameters:
    path (str | pathlib.Path): The path to the GeoJSON file.

    Returns:
    list[LayerDataTuple]: A list of LayerDataTuple containing the shape layer information.
    """

    if isinstance(path, str):
        path = pathlib.Path(path)

    geo_anno = geopandas.read_file(path)
    shape_layer_data = []
    for anno, _ in enumerate(geo_anno["name"]):
        print(anno, _)
        nap_anno = []
        if geo_anno["geometry"][anno].type == "Polygon":
            nap_anno.append(geo_anno["geometry"][anno].exterior.coords[:])
        elif geo_anno["geometry"][anno].type == "MultiPolygon":
            for poly in geo_anno["geometry"][anno].geoms:
                nap_anno.append(poly.exterior.coords[:])
        else:
            print(geo_anno["geometry"][anno].type)
            raise NotImplementedError(geo_anno["geometry"][anno].type)

        shape_layer_data.append(
            (
                nap_anno,
                {
                    "name": _,
                    "shape_type": "polygon",
                    "rotate": -90,
                    "scale": [
                        -1,
                        1,
                    ],  # Needs to be set to img layer scale, with *-1,*1
                    "metadata": {"from_geoJSON": True},
                },
                "shapes",
            )
        )
    return shape_layer_data


def load_parquet(path: str | pathlib.Path) -> list[LayerDataTuple]:
    """
    Load a Parquet file and return its contents as a list of LayerDataTuple.
    Parameters:
    path (str | pathlib.Path): The path to the Parquet file.
    Returns:
    list[LayerDataTuple]: A list containing the data from the Parquet file,
        formatted as a LayerDataTuple with the data as a NumPy array,
        an empty dictionary for metadata, and the string "labels".
    """
    print("WIP function")
    if isinstance(path, str):
        print("PATH IS A STRINGG!!!!!")
        path = pathlib.Path(path)

    df = pandas.read_parquet(path)
    print("Experiental")
    print(df)

    return [(df.to_numpy(), {}, "labels")]


# def save_geojson

# def wAnno(
#         viewer: "napari.Viewer",
#         imgAnno: "napari.layers.Image",
#         shapeAnno: "napari.layers.Shapes",
#         annoDir = pathlib.Path("")
#         ):

#     features = []
#     for layer in shapeAnno:
#         geoms = []
#         for shape in layer.data:
#             if layer.metadata.get("nap_generated"):
#                 shape = numpy.dot(shape, [[0, 1], [1, 0]]) # QUpath rotates and flips images. WIP to make it standard for our workflow too
#             geoms.append(Polygon(shape))
#         mpoly = MultiPolygon(geoms)
#         feature = {
#             "type": "Feature",
#             "properties": {
#                 "objectType":"annotation",
#                 "name":layer.name},
#                 "geometry": mpoly.__geo_interface__
#         }
#         features.append(feature)
#     feature_collection = {
#         "type":"FeatureCollection",
#         "features": features
#     }

#     gdf = geopandas.GeoDataFrame.from_features(feature_collection)
#     gdf.to_file(filename= annoDir / f"{imgAnno.name}.geojson", driver="GeoJSON")

# def save_parquet():
